/**
 * @file
 *
 * @ingroup rtems_media
 *
 * @brief Media implementation.
 */

/*
 * Copyright (c) 2009, 2010 embedded brains GmbH.  All rights reserved.
 *
 *  embedded brains GmbH
 *  Obere Lagerstr. 30
 *  82178 Puchheim
 *  Germany
 *  <rtems@embedded-brains.de>
 *
 * The license and distribution terms for this file may be
 * found in the file LICENSE in this distribution or at
 * http://www.rtems.com/license/LICENSE.
 */

#include <stdlib.h>

#include <sys/queue.h>

#include "media.h"

typedef struct rtems_media_listener {
  SLIST_ENTRY(rtems_media_listener) link;
  rtems_media_callback callback;
  void *arg;
} rtems_media_listener;

static SLIST_HEAD(rtems_media_listener_head, rtems_media_listener) rtems_media_listener_list =
  SLIST_HEAD_INITIALIZER(rtems_media_listener_list);

static rtems_media_listener *rtems_media_find_listener(rtems_media_callback callback, void *arg)
{
  rtems_media_listener *listener = NULL;

  SLIST_FOREACH(listener, &rtems_media_listener_list, link) {
    if (listener->callback == callback && listener->arg == arg) {
      return listener;
    }
  }

  return NULL;
}

rtems_status_code rtems_media_callback_add(rtems_media_callback callback, void *arg)
{
  rtems_media_listener *listener = rtems_media_find_listener(callback, arg);

  if (listener == NULL) {
    listener = malloc(sizeof(rtems_media_listener));
    if (listener != NULL) {
      listener->callback = callback;
      listener->arg = arg;
      SLIST_INSERT_HEAD(&rtems_media_listener_list, listener, link);
    } else {
      return RTEMS_NO_MEMORY;
    }
  } else {
    return RTEMS_TOO_MANY;
  }

  return RTEMS_SUCCESSFUL;
}

rtems_status_code rtems_media_callback_remove(rtems_media_callback callback, void *arg)
{
  rtems_media_listener *listener = rtems_media_find_listener(callback, arg);

  if (listener != NULL) {
    SLIST_REMOVE(&rtems_media_listener_list, listener, rtems_media_listener, link);
    free(listener);
  } else {
    return RTEMS_INVALID_ID;
  }

  return RTEMS_SUCCESSFUL;
}

rtems_status_code rtems_media_notify(rtems_media_event event, rtems_media_state state, const char *src, const char *dest)
{
  rtems_status_code sc = RTEMS_SUCCESSFUL;
  rtems_status_code rsc = RTEMS_SUCCESSFUL;
  rtems_media_listener *listener = NULL;

  SLIST_FOREACH(listener, &rtems_media_listener_list, link) {
    sc = (*listener->callback)(event, state, src, dest, listener->arg);
    if (sc != RTEMS_SUCCESSFUL) {
      rsc = sc;
    }
  }

  return rsc;
}
