/**
 * @file
 *
 * @ingroup rtems_media
 *
 * @brief Media API.
 */

/*
 * Copyright (c) 2009, 2010 embedded brains GmbH.  All rights reserved.
 *
 *  embedded brains GmbH
 *  Obere Lagerstr. 30
 *  82178 Puchheim
 *  Germany
 *  <rtems@embedded-brains.de>
 *
 * The license and distribution terms for this file may be
 * found in the file LICENSE in this distribution or at
 * http://www.rtems.com/license/LICENSE.
 */

#ifndef RTEMS_MEDIA_H
#define RTEMS_MEDIA_H

#include <rtems.h>
#include <rtems/bdpart.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define RTEMS_MEDIA_MOUNT_BASE "/media"

#define RTEMS_MEDIA_DELIMITER '-'

typedef enum {
  RTEMS_MEDIA_EVENT_DISK_ATTACH = 0,
  RTEMS_MEDIA_EVENT_DISK_DETACH,
  RTEMS_MEDIA_EVENT_MOUNT,
  RTEMS_MEDIA_EVENT_UNMOUNT,
  RTEMS_MEDIA_EVENT_PARTITION_INQUIRY,
  RTEMS_MEDIA_EVENT_PARTITION_ATTACH,
  RTEMS_MEDIA_EVENT_PARTITION_DETACH
} rtems_media_event;

typedef enum {
  RTEMS_MEDIA_STATE_INQUIRY,
  RTEMS_MEDIA_STATE_READY,
  RTEMS_MEDIA_STATE_ABORTED,
  RTEMS_MEDIA_STATE_SUCCESS,
  RTEMS_MEDIA_STATE_FAILED
} rtems_media_state;

typedef rtems_status_code (*rtems_media_callback)(rtems_media_event event, rtems_media_state state, const char *src, const char *dest, void *arg);

typedef rtems_status_code (*rtems_media_event_worker)(rtems_media_state state, const char *src, char **dest, void *arg);

rtems_status_code rtems_media_callback_add(rtems_media_callback callback, void *arg);

rtems_status_code rtems_media_callback_remove(rtems_media_callback callback, void *arg);

rtems_status_code rtems_media_initialize(void);

rtems_status_code rtems_media_server_initialize(
  rtems_task_priority priority,
  size_t stack_size,
  rtems_mode modes,
  rtems_attribute attributes,
  uint32_t message_count
);

rtems_status_code rtems_media_disk_detach(rtems_media_state state, const char *src, char **dest, void *arg);

rtems_status_code rtems_media_get_device_identifier(
  const char *device_path,
  dev_t *device_identifier
);

rtems_status_code rtems_media_notify(rtems_media_event event, rtems_media_state state, const char *src, const char *dest);

rtems_status_code rtems_media_post_event(rtems_media_event event, const char *src, char **dest, rtems_media_event_worker worker, void *arg);

rtems_status_code rtems_media_server_post_event(
  rtems_media_event event,
  const char *src,
  rtems_media_event_worker worker,
  void *arg
);

static inline rtems_status_code rtems_media_server_disk_attach(
  const char *disk_path,
  rtems_media_event_worker worker,
  void *arg
)
{
  return rtems_media_server_post_event(RTEMS_MEDIA_EVENT_DISK_ATTACH, disk_path, worker, arg);
}

static inline rtems_status_code rtems_media_server_disk_detach(const char *disk_path)
{
  return rtems_media_server_post_event(RTEMS_MEDIA_EVENT_DISK_DETACH, disk_path, NULL, NULL);
}

const char *rtems_media_event_assoc(rtems_media_event event);

const char *rtems_media_state_assoc(rtems_media_state state);

char *rtems_media_create_path(const char *prefix, const char *name, rtems_device_major_number major);

char *rtems_media_replace_prefix(const char *prefix, const char *path);

char *rtems_media_append_minor(const char *path, rtems_device_minor_number minor);

char *rtems_media_path_split(const char *path, const char **name_ptr);

#include <sys/stat.h>

typedef rtems_status_code (*rtems_per_directory_routine)(const char *entry_path, struct stat *st, void *arg);

rtems_status_code rtems_directory_iterate(
  const char *dir_path,
  rtems_per_directory_routine routine,
  void *routine_arg
);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* RTEMS_MEDIA_H */
